/* Revision History
 * 2015-02-07 - MisterCardholder
 *     The script can load models with UV maps, basic textures, armatures, and
 *     bone weights.
 *     It does not support vertex normals or more complex materials/shaders.
 * 2015-02-09 - MisterCardholder
 *     Added support for stage geometry (scr) files.
 */

/* Notes:
 *
 * The XBLA version of the game appears to use the same file format, only 
 * big-endian, but the assets look to be of a lower quality, so there 
 * probably isn't a good reason to extend this script to handle them.
 *
 * The vertex normals didn't look right in Noesis and seem to be ignored when
 * importing into Blender, so I haven't bothered debugging them.
 *
 * In one instance, this script was observed to not apply bone weights to
 * a subset of a mesh, so there may be some minor issues with the bone maps.
 *
 * A number of the models include extra meshes which are not visible during
 * gameplay.  Documentation on the Bayonetta models indicates that they may
 * be used to generate shadows.  By default, the script will ignore these
 * meshes, but they can be included in exports by using the
 * -wmb3includehiddenmeshes advanced option.
 *
 * The script will fail to load some Korra wmb files.  If the file is ~1KB in
 * size the issue is probably that the file doesn't contain any vertexes.  If
 * the file is larger, it's possible that it's using using a vertexStride or
 * boneWeightStride that isn't being handled, but the more likely explanation
 * is that it's one of a handful of WMB4 files that Korra uses.
 *
 * When creating materials, the script currently ignores everything except for
 * the albedo map.  Some of the other texture types, like the specular map,
 * appear to be included for most models, but others (like the cube map) won't
 * be found in the corresponding wta/wtp file, so you'd need to go seek them
 * out if you wanted to support them.
 *
 * When exporting stages, Noesis seems to export one model at a time, but 
 * includes <numberOfModelsInStage> copies of each texture, whether they're 
 * used in that particular model or not.  I'm not sure if that's because I'm
 * doing something wrong or if that's just how Noesis works when dealing with
 * multiple models.
 *
 */

#include "stdafx.h"
#include <stddef.h>
#include "wmb3.h"
#include "wta.h"
#include "scr.h"

typedef struct wmb3Opts_s {
	bool includeHiddenMeshes;
} wmb3Opts_t;

const char *g_pPluginName = "TheLegendOfKorraPC";
const char *g_pPluginDesc = "The Legend of Korra (PC) model handler, by MisterCardholder";

wmb3Opts_t *g_opts = NULL;

/// Determines if we can handle the given file
bool korraTypeCheckModel(BYTE *fileBuffer, int bufferLen, noeRAPI_t *rapi) {
	if (bufferLen < sizeof(wmb3_header_t)) {
		// Too small to be a WMB file
		return false;
	}
	wmb3_header_t *fileHeader = ((wmb3_header_t *)fileBuffer);
	if (memcmp(&(fileHeader->fourCC), "WMB3", 4)) {
		// WMB3 header missing
		return false;
	}
	if (fileHeader->unknown04 != 0x00010000) {
		// Magic number is wrong (could be an endian-ness issue)
		return false;
	}
	// All signs point to a valid file
	return true;
}

/// Determines if we can handle the given file
bool korraTypeCheckStage(BYTE *fileBuffer, int bufferLen, noeRAPI_t *rapi) {
	if (bufferLen < sizeof(scr_header_t)) {
		// Too small to be a WMB file
		return false;
	}
	scr_header_t *fileHeader = ((scr_header_t *)fileBuffer);
	if (memcmp(&(fileHeader->fourCC), "SCR\0", 4)) {
		// SCR header missing
		return false;
	}
	// All signs point to a valid file
	return true;
}

void unpackNormals(uint32_t packed, float *x, float *y, float *z) {
	int xWidth = 11;
	int yWidth = 11;
	int zWidth = 10;
	int xPos = 0;
	int yPos = xPos + xWidth;
	int zPos = yPos + yWidth;
	size_t fullWidth = sizeof(packed) * 8;

	int xInt = ((int)(packed << (fullWidth - xPos - xWidth))) >> (fullWidth - xWidth);
	int yInt = ((int)(packed << (fullWidth - yPos - yWidth))) >> (fullWidth - yWidth);
	int zInt = ((int)(packed << (fullWidth - zPos - zWidth))) >> (fullWidth - zWidth);

	*x = (xInt) / (float)((1 << (xWidth - 1)) - 1);
	*y = (yInt) / (float)((1 << (yWidth - 1)) - 1);
	*z = (zInt) / (float)((1 << (zWidth - 1)) - 1);
}

// You must free the pointer returned by Noesis_ReadFile with Noesis_UnpooledFree! (unless it's NULL)
BYTE *korraLoadRelatedFile(char *extension, noeRAPI_t *rapi, int &filesize) {
	char *inFileName = rapi->Noesis_GetInputName();
	BYTE *relatedFile = NULL;
	filesize = 0;
	if (inFileName && inFileName[0]) {
		char relatedFileName[MAX_NOESIS_PATH];
		rapi->Noesis_GetExtensionlessName(relatedFileName, inFileName);
		if (rapi->Noesis_CheckFileExt(inFileName, ".scr")) {
			// The naming scheme for wta/wtp files associated with scr files
			// is slightly different: xxx.scr -> xxxscr.wta NOT xxx.wta
			strcat_s(relatedFileName, MAX_NOESIS_PATH, "scr");
		}
		strcat_s(relatedFileName, MAX_NOESIS_PATH, extension);

		int relatedFileLength = 0;
		relatedFile = (BYTE *)rapi->Noesis_ReadFile(relatedFileName, &relatedFileLength);
		filesize = relatedFileLength;
	}
	return relatedFile;
}

bool korraValidateWTAFile(BYTE *fileBuffer, int bufferLen) {
	if (bufferLen < sizeof(wta_header_t)) {
		// Too small to be a WTA file
		return false;
	}
	wta_header_t *fileHeader = ((wta_header_t *)fileBuffer);
	if (memcmp(&(fileHeader->fourCC), "WTB\0", 4)) {
		// WTA header missing
		return false;
	}
	return true;
}

void korraLoadTextures(CArrayList<noesisTex_t *> &textures, noeRAPI_t *rapi) {
	int wtaFileLength = 0;
	BYTE *wtaFile = NULL;
	int wtpFileLength = 0;
	BYTE *wtpFile = NULL;

	wtaFile = korraLoadRelatedFile(".wta", rapi, wtaFileLength);

	if (korraValidateWTAFile(wtaFile, wtaFileLength)) {
		wta_header_t *wtaHeader = (wta_header_t *)wtaFile;
		uint32_t textureCount = wtaHeader->textureCount;
		wta_textureOffset_t *textureOffsetBuffer = (wta_textureOffset_t *)(wtaFile + wtaHeader->textureOffsetArrayOffset);
		wta_textureSize_t *textureSizeBuffer = (wta_textureSize_t *)(wtaFile + wtaHeader->textureSizeArrayOffset);
		wta_textureIdentifier_t *textureIdentifierBuffer = (wta_textureIdentifier_t *)(wtaFile + wtaHeader->textureIdentifierArrayOffset);

		wtpFile = korraLoadRelatedFile(".wtp", rapi, wtpFileLength);

		for (int i = 0; i < textureCount; i++) {
			noesisTex_t *texture = NULL;
			if ((textureOffsetBuffer[i] + textureSizeBuffer[i]) <= wtpFileLength) {
				texture = rapi->Noesis_LoadTexByHandler((wtpFile + textureOffsetBuffer[i]), textureSizeBuffer[i], ".dds");
			}
			if (!texture) {
				texture = rapi->Noesis_AllocPlaceholderTex(NULL, 32, 32, false);
				rapi->LogOutput("WARNING: Texture %d not found in the WTP file.\n", i);
			}
			texture->globalIdx = textureIdentifierBuffer[i];
			textures.Append(texture);
		}
	}

	if (wtaFile) {
		rapi->Noesis_UnpooledFree(wtaFile);
	}
	if (wtpFile) {
		rapi->Noesis_UnpooledFree(wtpFile);
	}
}

// Searches the list of textures for one with the given identifier
// Returns -1 if one cannot be found
int korraIndexOfTextureWithIdentifier(CArrayList<noesisTex_t *> &textures, uint32_t textureIdentifier) {
	int textureIndex = -1;
	for (int k = 0; k < textures.Num(); k++) {
		noesisTex_t *tex = textures[k];
		if (tex && (tex->globalIdx == textureIdentifier)) {
			textureIndex = k;
			break;
		}
	}
	return textureIndex;
}

void korraCreateMaterials(wmb3_header_t &fileHeader, BYTE *fileBuffer, CArrayList<noesisMaterial_t *> &materials, CArrayList<noesisTex_t *> &textures, noeRAPI_t *rapi) {
	wmb3_textureIdentifier_t *textureIdentifierBuffer = (wmb3_textureIdentifier_t *)(fileBuffer + fileHeader.textureIdentifierArrayOffset);
	uint32_t textureIdentifierCount = fileHeader.textureIdentifierCount;

	wmb3_material_t *materialBuffer = (wmb3_material_t *)(fileBuffer + fileHeader.materialArrayOffset);
	uint32_t materialCount = fileHeader.materialCount;

	// For each material...
	for (int i = 0; i < materialCount; i++) {
		noesisMaterial_t *noesisMaterial = rapi->Noesis_GetMaterialList(1, false);
		noesisMaterial->texIdx = -1;

		wmb3_textureMap_t *textureMapBuffer = (wmb3_textureMap_t *)(fileBuffer + materialBuffer[i].textureMapArrayOffset);
		uint32_t textureMapCount = materialBuffer[i].textureMapCount;

		// For each texture assignment in each material...
		for (int j = 0; j < textureMapCount; j++) {

			// If we haven't yet found an albedo map...
			if ((noesisMaterial->texIdx == -1) && !memcmp((fileBuffer + textureMapBuffer[j].textureTypeNameOffset), "Albedo", 6)) {
				uint32_t textureIdentifierIndex = textureMapBuffer[j].textureIdentifierIndex;

				int textureIndex = -1;

				// Search each loaded texture for one with a matching identifier
				if (textureIdentifierIndex < textureIdentifierCount) {
					uint32_t textureIdentifier = textureIdentifierBuffer[textureIdentifierIndex];
					textureIndex = korraIndexOfTextureWithIdentifier(textures, textureIdentifier);
				}

				noesisMaterial->texIdx = textureIndex;
			}
		}

		if (noesisMaterial->texIdx < 0) {
			// The requested texture wasn't found, so insert a placeholder
			noesisTex_t *noesisTexture = rapi->Noesis_AllocPlaceholderTex(NULL, 32, 32, false);
			noesisMaterial->texIdx = textures.Num();
			textures.Append(noesisTexture);
			rapi->LogOutput("WARNING: Failed to load an albedo texture for material %d.\n", i);
		}

		char matName[128];
		sprintf_s(matName, 128, "material%03d", i);
		noesisMaterial->name = rapi->Noesis_PooledString(matName);

		// Without this next line, parts of the model will appear transparent until you cycle the blend mode
		// With it, parts that SHOULD be transparent (like pro-bending masks) won't be and you won't be able to cycle the blend mode
		//noesisMaterial->noDefaultBlend = true;

		materials.Append(noesisMaterial);
	}
}

modelBone_t *korraCreateBones(wmb3_header_t &fileHeader, BYTE *fileBuffer, noeRAPI_t *rapi, int &numBones) {
	numBones = 0;
	if (fileHeader.boneCount == 0) {
		return NULL;
	}
	wmb3_bone_t *boneBuffer = (wmb3_bone_t *)(fileBuffer + fileHeader.boneArrayOffset);
	numBones = fileHeader.boneCount;

	modelBone_t *bones = rapi->Noesis_AllocBones(numBones);

	for (int i = 0; i < numBones; i++) {
		modelBone_t *bone = bones + i;

		sprintf_s(bone->name, 30, "bone%03d", boneBuffer[i].boneNumber);

		short parent = boneBuffer[i].parentIndex;
		assert(parent < numBones);
		bone->eData.parent = (parent >= 0) ? &(bones[parent]) : NULL;

		bone->mat = g_identityMatrix;
		float pos[3];
		memcpy(pos, &(boneBuffer[i].positionAbsoluteX), sizeof(pos));
		g_mfn->Math_VecCopy(pos, bone->mat.o);
	}

	return bones;
}

noesisModel_t *korraCreateModel(BYTE *fileBuffer, CArrayList<noesisTex_t *> &textures, noeRAPI_t *rapi) {

	wmb3_header_t *fileHeader = (wmb3_header_t *)fileBuffer;

	wmb3_vertexHeader_t *vertexHeaderBuffer = (wmb3_vertexHeader_t *)(fileBuffer + fileHeader->vertexHeaderArrayOffset);
	
	void *vertexBuffer = (fileBuffer + vertexHeaderBuffer->vertexArrayOffset);
	void *boneWeightBuffer = (fileBuffer + vertexHeaderBuffer->boneWeightArrayOffset);
	uint32_t vertexCount = vertexHeaderBuffer->vertexCount;

	uint32_t vertexStride;
	int32_t vertexPositionOffset;
	int32_t vertexUVOffset;
	int32_t vertexNormalsOffset;
	switch(vertexHeaderBuffer->vertexStride) {
		case 20:
			vertexStride 			= sizeof(wmb3_vertex20_t);
			vertexPositionOffset 	= offsetof(wmb3_vertex20_s, positionX);
			vertexUVOffset 			= offsetof(wmb3_vertex20_s, textureU);
			vertexNormalsOffset 	= offsetof(wmb3_vertex20_s, normals);
			break;
		case 24:
			vertexStride 			= sizeof(wmb3_vertex24_t);
			vertexPositionOffset 	= offsetof(wmb3_vertex24_s, positionX);
			vertexUVOffset 			= offsetof(wmb3_vertex24_s, textureU);
			vertexNormalsOffset 	= offsetof(wmb3_vertex24_s, normals);
			break;
		case 28:
			vertexStride 			= sizeof(wmb3_vertex28_t);
			vertexPositionOffset 	= offsetof(wmb3_vertex28_t, positionX);
			vertexUVOffset 			= offsetof(wmb3_vertex28_t, textureU);
			vertexNormalsOffset 	= offsetof(wmb3_vertex28_t, normals);
			break;
		default:
			// Unknown vertex format
			return NULL;
	}

	uint32_t boneWeightStride;
	int32_t boneIndexOffset;
	int32_t boneWeightOffset;
	switch (vertexHeaderBuffer->boneWeightStride) {
		case 24:
			boneWeightStride = sizeof(wmb3_boneWeight24_t);
			boneIndexOffset = offsetof(wmb3_boneWeight24_s, boneIndex0);
			boneWeightOffset = offsetof(wmb3_boneWeight24_s, boneWeight0);
			break;
		case 20:
			boneWeightStride = sizeof(wmb3_boneWeight20_t);
			boneIndexOffset = offsetof(wmb3_boneWeight20_s, boneIndex0);
			boneWeightOffset = offsetof(wmb3_boneWeight20_s, boneWeight0);
			break;
		case 16:
			boneWeightStride = sizeof(wmb3_boneWeight16_t);
			boneIndexOffset = offsetof(wmb3_boneWeight16_s, boneIndex0);
			boneWeightOffset = offsetof(wmb3_boneWeight16_s, boneWeight0);
			break;
		case 8:
			boneWeightStride = sizeof(wmb3_boneWeight8_t);
			boneIndexOffset = -1;
			boneWeightOffset = -1;
			break;
		default:
			// Unknown bone weight format
			return NULL;
	}

	wmb3_face_t *faceBuffer = (wmb3_face_t *)(fileBuffer + vertexHeaderBuffer->faceArrayOffset);
	uint32_t faceCount = vertexHeaderBuffer->faceCount;

	wmb3_mesh_t *meshBuffer = (wmb3_mesh_t *)(fileBuffer + fileHeader->meshArrayOffset);
	uint32_t meshCount = fileHeader->meshCount;

	wmb3_meshMetadataHeader_t *meshMetadataHeader = (wmb3_meshMetadataHeader_t *)(fileBuffer + fileHeader->meshMetadataHeaderOffset);
	wmb3_meshMetadata_t *meshMetadataBuffer = (wmb3_meshMetadata_t *)(fileBuffer + meshMetadataHeader->meshMetadataOffset);
	uint32_t meshMetadataCount = meshMetadataHeader->meshMetadataCount;
	wmb3_meshMetadata_t *hiddenMeshMetadataBuffer = (wmb3_meshMetadata_t *)(fileBuffer + meshMetadataHeader->hiddenMeshMetadataOffset);
	uint32_t hiddenMeshMetadataCount = meshMetadataHeader->hiddenMeshMetadataCount;

	wmb3_meshNameHeader_t *meshNameHeaderBuffer = (wmb3_meshNameHeader_t *)(fileBuffer + fileHeader->meshNameHeaderArrayOffset);
	uint32_t meshNameHeaderCount = fileHeader->meshNameHeaderCount;

	wmb3_boneMap_t *boneMapBuffer = (wmb3_boneMap_t *)(fileBuffer + fileHeader->boneMapOffset);
	uint32_t boneMapCount = fileHeader->boneMapCount;

	CArrayList<noesisMaterial_t *> materials;
	korraCreateMaterials(*fileHeader, fileBuffer, materials, textures, rapi);

	int numBones;
	modelBone_t *bones = korraCreateBones(*fileHeader, fileBuffer, rapi, numBones);


	void *pgctx = rapi->rpgCreateContext();

	// For each mesh...
	for (int i = 0; i < meshCount; i++) {
		// Find the mesh's metadata
		wmb3_meshMetadata_t *metadata = NULL;
		bool isHiddenMesh;
		for (int j = 0; j < meshMetadataCount; j++) {
			if ((meshMetadataBuffer + j)->meshIndex == i) {
				metadata = (meshMetadataBuffer + j);
				isHiddenMesh = false;
				break;
			}
		}
		if (!metadata && g_opts->includeHiddenMeshes) {
			for (int j = 0; j < hiddenMeshMetadataCount; j++) {
				if ((hiddenMeshMetadataBuffer + j)->meshIndex == i) {
					metadata = (hiddenMeshMetadataBuffer + j);
					isHiddenMesh = true;
					break;
				}
			}
		}
		if (metadata == NULL) {
			// Ignore meshes without metadata
			continue;
		}

		// Set the mesh's name
		uint32_t meshNameIndex = metadata->meshNameIndex;
		if (!isHiddenMesh && (meshNameIndex < meshNameHeaderCount)) {
			char meshName[64];
			sprintf(meshName, "mesh%02d-%s", i, (char *)(fileBuffer + meshNameHeaderBuffer[meshNameIndex].nameOffset));
			rapi->rpgSetName(meshName);
		} else {
			char meshName[64];
			sprintf(meshName, "mesh%02d", i);
			rapi->rpgSetName(meshName);
		}

		// Get the bone reference map
		int *boneReferenceMap = NULL;
		uint32_t boneMapIndex = metadata->boneMapIndex;
		if (boneMapIndex < boneMapCount) {
			wmb3_boneMap_t *boneMap = &(boneMapBuffer[boneMapIndex]);
			wmb3_boneMapEntry_t *boneMapEntries = (wmb3_boneMapEntry_t *)(fileBuffer + boneMap->boneMapOffset);
			boneReferenceMap = (int *)rapi->Noesis_UnpooledAlloc(sizeof(int)*boneMap->boneMapCount);
			for (int j = 0; j < boneMap->boneMapCount; j++) {
				boneReferenceMap[j] = boneMapEntries[j];
			}
			rapi->rpgSetBoneMap(boneReferenceMap);
		}

		// Set the material
		if (!isHiddenMesh) {
			uint32_t materialIndex = metadata->materialIndex;
			rapi->rpgSetMaterialIndex(materialIndex);
		}

		rapi->rpgBegin(RPGEO_TRIANGLE);

		// For each (triangular) face...
		for (int j = 0; j < meshBuffer[i].faceCount; j += 3) {
			int faceBufferIndex = meshBuffer[i].faceStart + j;
			int vertexStartIndex = meshBuffer[i].vertexStart;	// The first vertex in this mesh

			// For each vertex...
			for (int k = 0; k < 3; k++) {
				int vertexBufferIndex = vertexStartIndex + faceBuffer[faceBufferIndex + k];

				if ((boneIndexOffset >= 0) && (boneWeightOffset >= 0)) {
					rapi->rpgVertBoneIndexUB((BYTE *)((uint8_t *)boneWeightBuffer + boneIndexOffset + (vertexBufferIndex * boneWeightStride)), 4);
					rapi->rpgVertBoneWeightUB((BYTE *)((uint8_t *)boneWeightBuffer + boneWeightOffset + (vertexBufferIndex * boneWeightStride)), 4);
				}

				/* Normals don't look quite right.  There may be an issue with the unpacking function.
				uint32_t packedNormals = *((float *)((uint8_t *)vertexBuffer + vertexNormalsOffset + (vertexBufferIndex * vertexStride)));
				float normals[3];
				unpackNormals(packedNormals, normals, normals + 1, normals + 2);
				rapi->rpgVertNormal3f(normals);
				*/

				rapi->rpgVertUV2hf((WORD *)((uint8_t *)vertexBuffer + vertexUVOffset + (vertexBufferIndex * vertexStride)), 0);
				rapi->rpgVertex3f((float *)((uint8_t *)vertexBuffer + vertexPositionOffset + (vertexBufferIndex * vertexStride)));
			}
		}

		rapi->rpgEnd();

		if (boneReferenceMap) {
			rapi->rpgSetBoneMap(NULL);
			rapi->Noesis_UnpooledFree(boneReferenceMap);
		}

		rapi->rpgSetMaterialIndex(-1);
	}

	noesisMatData_t *md = rapi->Noesis_GetMatDataFromLists(materials, textures);
	rapi->rpgSetExData_Materials(md);

	rapi->rpgSetExData_Bones(bones, numBones);

	noesisModel_t *mdl = rapi->rpgConstructModel();

	rapi->rpgDestroyContext(pgctx);

	return mdl;
}

noesisModel_t *korraLoadModel(BYTE *fileBuffer, int bufferLen, int &numMdl, noeRAPI_t *rapi) {
	CArrayList<noesisTex_t *> textures;
	korraLoadTextures(textures, rapi);

	noesisModel_t *mdl = korraCreateModel(fileBuffer, textures, rapi);
	if (mdl) {
		numMdl = 1;
	}

	return mdl;
}

noesisModel_t *korraLoadStage(BYTE *fileBuffer, int bufferLen, int &numMdl, noeRAPI_t *rapi) {
	scr_header_t *fileHeader = (scr_header_t *)fileBuffer;

	scr_modelHeaderOffset_t *modelHeaderOffsetBuffer = (scr_modelHeaderOffset_t *)(fileBuffer + fileHeader->modelHeaderOffsetArrayOffset);
	uint32_t modelHeaderCount = fileHeader->modelHeaderCount;
	
	CArrayList<noesisTex_t *> textures;
	korraLoadTextures(textures, rapi);

	CArrayList<noesisModel_t *> models;
	for (int i = 0; i < modelHeaderCount; i++) {
		scr_modelHeader_t *modelHeader = (scr_modelHeader_t *)(fileBuffer + modelHeaderOffsetBuffer[i]);
		BYTE *wmbBuffer = fileBuffer + modelHeader->wmbOffset;

		noesisModel_t *mdl = korraCreateModel(wmbBuffer, textures, rapi);
		if (mdl) {
			models.Append(mdl);
		}
	}

	if (models.Num() <= 0) {
		return NULL;
	}
	noesisModel_t *mdlList = rapi->Noesis_ModelsFromList(models, numMdl);
	models.Clear();

	return mdlList;
}

//handle -wmb3includehiddenmeshes
static bool Model_WMB3_OptHandler_IncludeHiddenMeshes(const char *arg, unsigned char *store, int storeSize) {
	wmb3Opts_t *lopts = (wmb3Opts_t *)store;
	assert(storeSize == sizeof(wmb3Opts_t));
	lopts->includeHiddenMeshes = true;
	return true;
}

//called by Noesis to init the plugin
bool NPAPI_InitLocal(void) {
	if (g_nfn->NPAPI_GetAPIVersion() < NOESIS_PLUGINAPI_VERSION) {
		return false;
	}

	int fh = g_nfn->NPAPI_Register("The Legend of Korra (PC) Model", ".wmb");
	if (fh < 0)	{
		return false;
	}

	//set the data handlers for this format
	g_nfn->NPAPI_SetTypeHandler_TypeCheck(fh, korraTypeCheckModel);
	g_nfn->NPAPI_SetTypeHandler_LoadModel(fh, korraLoadModel);

	//add first parm
	addOptParms_t optParms;
	memset(&optParms, 0, sizeof(optParms));
	optParms.optName = "-wmb3includehiddenmeshes";
	optParms.optDescr = "Include hidden meshes";
	optParms.storeSize = sizeof(wmb3Opts_t);
	optParms.handler = Model_WMB3_OptHandler_IncludeHiddenMeshes;
	g_opts = (wmb3Opts_t *)g_nfn->NPAPI_AddTypeOption(fh, &optParms);
	assert(g_opts);
	optParms.shareStore = (unsigned char *)g_opts;


	fh = g_nfn->NPAPI_Register("The Legend of Korra (PC) Stage", ".scr");
	if (fh < 0)	{
		return false;
	}

	//set the data handlers for this format
	g_nfn->NPAPI_SetTypeHandler_TypeCheck(fh, korraTypeCheckStage);
	g_nfn->NPAPI_SetTypeHandler_LoadModel(fh, korraLoadStage);

	return true;
}

//called by Noesis before the plugin is freed
void NPAPI_ShutdownLocal(void) {
	// nothing to do here
}

BOOL APIENTRY DllMain(HMODULE hModule, DWORD  ul_reason_for_call, LPVOID lpReserved) {
	return TRUE;
}
