/* Revision History
 * 2015-02-07 - MisterCardholder
 *   Initial Release
 */

/*
 * The Legend of Korra (PC) WMB File Structure
 *
 * +-------------------------------------+
 * | File Header                         |
 * | wmb3_header_t                       |
 * +-------------------------------------+
 * | Array of Vertex Headers             |
 * | wmb3_vertexHeader_t                 |
 * |   ----------------------------------+
 * |   Array of Vertexes                 |
 * |   wmb3_vertexXX_t                   |
 * |   ----------------------------------+
 * |   Array of Bone Weights             |
 * |   wmb3_boneWeightXX_t               |
 * |   ----------------------------------+
 * |   Array of Faces                    |
 * |   wmb3_face_t                       |
 * +-------------------------------------+
 * | Array of Meshes                     |
 * | wmb3_mesh_t                         |
 * +-------------------------------------+
 * | Mesh Metadata Header                |
 * | wmb3_meshMetadataHeader_t           |
 * |   ----------------------------------+
 * |   Array of Visible Mesh Metadata    |
 * |   wmb3_meshMetadata_t               |
 * |   ----------------------------------+
 * |   Array of Hidden Mesh Metadata     |
 * |   wmb3_meshMetadata_t               |
 * +-------------------------------------+
 * | Array of Bones                      |
 * | wmb3_bone_t                         |
 * +-------------------------------------+
 * | Array of Unknown Data               |
 * | Appears to be an array of uint8_t's |
 * | I've seen a suggestion that they    |
 * | might be normalized floats (so      |
 * | divide by 0xFF to get a value [0,1])|
 * | but their purpose is unclear.       |
 * +-------------------------------------+
 * | Array of Bone Map Offsets           |
 * | wmb3_boneMap_t                      |
 * |   ----------------------------------+
 * |   Array of Bone Map Indexes         |
 * |   uint16_t                          |
 * +-------------------------------------+
 * | Array of Materials                  |
 * | wmb3_material_t                     |
 * |   ----------------------------------+
 * |   Materials Data                    |
 * |   wmb3_textureMap_t                 |
 * +-------------------------------------+
 * | Array of Texture Identifiers        |
 * | wmb3_textureIdentifier_t            |
 * +-------------------------------------+
 * | Array of Mesh Names Headers         |
 * | wmb3_meshNameHeader_t               |
 * |   ----------------------------------+
 * |   Mesh Name Data                    |
 * |                                     |
 * +-------------------------------------+
 *
 * Notes:
 * 
 * All offsets are from the start of the file.
 *
 * Many sections of the file are zero-padded to be 16-byte-aligned, so
 * there may be instances where structures either mistakenly include those 
 * zeros or omit fields because they happened to be zero when documenting the
 * file structure.
 *
 */

#include <stdint.h>

typedef struct wmb3_header_s {
	char		fourCC[4]; 	// "WMB3"
	uint32_t	unknown04;	// Always 65,536
	uint32_t	unknown08;	// 0x00000000 (PC) 0xFFFFFFFF (360)
	uint32_t	unknown0C;	// 0x010137 or 0x000137 or 0x000107 (bit fields of some sort?)
	uint16_t	unknown10;	// 0 (PC) 1 (360)
	uint16_t	unknown12;	// usually 0xFFFF (but sometimes 0x0000)
	float		unknown14;
	float		unknown18;
	float		unknown1C;
	float		unknown20;
	float		unknown24;
	float		unknown28;
	uint32_t	vertexHeaderArrayOffset;		// 0x2C see: wmb3_vertexHeader_t
	uint32_t	vertexHeaderArrayCount;
	uint32_t	meshArrayOffset;				// 0x34 see: wmb3_mesh_t
	uint32_t	meshCount;
	uint32_t	meshMetadataHeaderOffset;		// 0x3C see: wmb3_meshMetadataHeader_t
	uint32_t	boneArrayOffset;				// 0x40 see: wmb3_bone_t
	uint32_t	boneCount;
	uint32_t	unknown48;						// Offset
	uint32_t	unknown4C;						// Count
	uint32_t	boneMapOffset;					// 0x50 see wmb3_boneMap_t
	uint32_t	boneMapCount;
	uint32_t	materialArrayOffset;			// 0x58 see wmb3_material_t
	uint32_t	materialCount;
	uint32_t	textureIdentifierArrayOffset;	// 0x60 see wmb3_textureIdentifier_t
	uint32_t	textureIdentifierCount;
	uint32_t	meshNameHeaderArrayOffset;		// 0x68 see: wmb3_meshNameHeader_t
	uint32_t	meshNameHeaderCount;
	uint32_t	unknown70;						// +
	uint32_t	unknown74;						// | Always 0x00000000
	uint32_t	unknown78;						// | They might just be padding
	uint32_t	unknown7C;						// +
} wmb3_header_t;

/********** Vertex Data **********/

typedef struct wmb3_vertexHeader_s {
	uint32_t	vertexArrayOffset;		// see: wmb3_vertexXX_t
	uint32_t	boneWeightArrayOffset;	// see: wmb3_boneWeightXX_t
	uint32_t	unknown08;	// 0
	uint32_t	unknown0C;	// 0
	uint32_t	vertexStride;		// Determines the number of bytes per vertex
	uint32_t	boneWeightStride;	// Determines the number of bytes per bone weight
	uint32_t	unknown18;	// 0
	uint32_t	unknown1C;	// 0
	uint32_t	vertexCount;		// Applies to both vertexArrayOffset and boneWeightArrayOffset
	uint32_t	faceArrayOffset;	// see: wmb3_face_t
	uint32_t	faceCount;
} wmb3_vertexHeader_t;

// Note: The normals appear to be packed in a 10.11.11 format as in the
// OpenGL Spec https://www.opengl.org/registry/doc/glspec45.core.pdf
// but this is currently unverified.
typedef struct wmb3_vertex20_s {
	float		positionX;
	float		positionY;
	float		positionZ;
	uint16_t	textureU;	// Actually a half-precision float
	uint16_t	textureV;	// Actually a half-precision float
	uint32_t	normals;
} wmb3_vertex20_t;

typedef struct wmb3_vertex24_s {
	float		positionX;
	float		positionY;
	float		positionZ;
	uint16_t	textureU;	// Actually a half-precision float
	uint16_t	textureV;	// Actually a half-precision float
	uint32_t	normals;
	uint32_t	unknown24;	// 0xFFFFFFFF
} wmb3_vertex24_t;

typedef struct wmb3_vertex28_s {
	float		positionX;
	float		positionY;
	float		positionZ;
	uint16_t	textureU;	// Actually a half-precision float
	uint16_t	textureV;	// Actually a half-precision float
	uint32_t	normals;
	uint32_t	unknown24;	// 0xFFFFFFFF
	uint32_t	unknown28;	// Possibly two half-precision floats?
} wmb3_vertex28_t;


// The boneIndex values in the wmb3_boneWeightXX_t structs are indexes for the
// relevant drawcall/mesh, NOT for the global list of bones. The per-mesh bone
// maps are stored as wmb3_boneMap_t.
typedef struct wmb3_boneWeight8_s {
	uint32_t	unknown00;
	uint32_t	unknown04;
} wmb3_boneWeight8_t;

typedef struct wmb3_boneWeight16_s {
	uint32_t	unknown00;
	uint32_t	unknown04;
	uint8_t		boneIndex0;		// +
	uint8_t		boneIndex1;		// | The index of the connected bone
	uint8_t		boneIndex2;		// |
	uint8_t		boneIndex3;		// +
	uint8_t		boneWeight0;	// +
	uint8_t		boneWeight1;	// | The weight of the connected bone
	uint8_t		boneWeight2;	// |
	uint8_t		boneWeight3;	// +
} wmb3_boneWeight16_t;

typedef struct wmb3_boneWeight20_s {
	uint32_t	unknown00;
	uint32_t	unknown04;
	uint8_t		boneIndex0;		// +
	uint8_t		boneIndex1;		// | The index of the connected bone
	uint8_t		boneIndex2;		// |
	uint8_t		boneIndex3;		// +
	uint8_t		boneWeight0;	// +
	uint8_t		boneWeight1;	// | The weight of the connected bone
	uint8_t		boneWeight2;	// |
	uint8_t		boneWeight3;	// +
	uint32_t	unknown20;		// 0xFFFFFFFF
} wmb3_boneWeight20_t;

typedef struct wmb3_boneWeight24_s {
	uint32_t	unknown00;
	uint32_t	unknown04;
	uint8_t		boneIndex0;		// +
	uint8_t		boneIndex1;		// | The index of the connected bone
	uint8_t		boneIndex2;		// |
	uint8_t		boneIndex3;		// +
	uint8_t		boneWeight0;	// +
	uint8_t		boneWeight1;	// | The weight of the connected bone
	uint8_t		boneWeight2;	// |
	uint8_t		boneWeight3;	// +
	uint32_t	unknown20;		// 0xFFFFFFFF
	uint32_t	unknown24;
} wmb3_boneWeight24_t;

// The vertex indexes used for the faces are based on the first vertex
// in the mesh, NOT in the global vertex array.
// If you don't offset by the mesh's vertexStart you're
// going to have a bad time.
//
// Although all of the models appear to use tris, the faces don't group
// verticies in threes, so the format seems like it could support quads
// or n-gons too.
typedef uint16_t wmb3_face_t;

/********** Mesh Data **********/

typedef struct wmb3_mesh_s {
	uint32_t	unknown00;		// 0x0000
	uint32_t	vertexStart;	// The index of the first vertex in this mesh
	uint32_t	faceStart;		// The first face index in this mesh
	uint32_t	vertexCount;	// The number of verticies in the mesh
	uint32_t	faceCount;		// The number of face indexes in the mesh
} wmb3_mesh_t;

typedef struct wmb3_meshMetadataHeader_s {
	uint32_t	meshMetadataOffset;		// see: wmb3_meshMetadata_t
	uint32_t	meshMetadataCount;
	uint32_t	unknown08;				// 0x0000
	uint32_t	unknown0C;				// 0x0000
	uint32_t	unknown10;				// 0x0000
	uint32_t	unknown14;				// 0x0000
	uint32_t	hiddenMeshMetadataOffset;	// see: wmb3_meshMetadata_t (applies to a mesh that's not part of the visible model; possibly a hitbox or similar piece of simple geometry)
	uint32_t	hiddenMeshMetadataCount;
	// Possibly up to 5 more 32-bit fields, but they're all zero, so I can't tell if they might just be padding
} wmb3_meshMetadataHeader_t;

typedef struct wmb3_meshMetadata_s {
	uint32_t	meshIndex;
	uint32_t	meshNameIndex;	// The index of the wmb3_meshName_t that applies to this mesh
	uint16_t	materialIndex;	// The index of the wmb3_material_t that applies to this mesh
	uint16_t	boneMapIndex;	// The index of the wmb3_boneMap_t that applies to this mesh
	uint32_t	unknown0C;		// Possibly an index, but it always seems to be 1
} wmb3_meshMetadata_t;

/********** Bone Data **********/

typedef struct wmb3_bone_s {
	uint16_t	unknown00;
	uint16_t	boneNumber;			// The bone's number (Display only! Don't use this for computing indexes or offsets or whatever)
	uint16_t	parentIndex;		// Index of the bone's parent in the global bone array (-1 if the bone has no parent)
	uint16_t	unknown06;			// Almost always 0x00
	float		positionRelativeX;	// + 
	float		positionRelativeY;	// | The bone's position relative to its parent (or the origin if it doesn't have a parent)
	float		positionRelativeZ;	// + 
	float		positionAbsoluteX;
	float		positionAbsoluteY;
	float		positionAbsoluteZ;
} wmb3_bone_t;

typedef struct wmb3_boneMap_s {
	uint32_t	boneMapOffset;		// see: wmb3_boneMapEntry_t
	uint32_t	boneMapCount;
} wmb3_boneMap_t;

// The index of the bone being mapped in the global bone array
typedef uint8_t wmb3_boneMapEntry_t;

/********** Materials Data **********/

typedef struct wmb3_material_s {
	uint32_t	shaderNameOffset;		// Offset to a null-terminated string that may be the name of the shader to use (observed values: "ois00_sbxxX", "cnm20_SbXXX")
	uint32_t	textureMapArrayOffset;	// see: wmb3_textureMap_t
	uint32_t	unknown08;				// 0x00000000
	uint32_t	matrixOffset;			// Offset to an array of floats; appears to be a matrix
	uint16_t	unknown10;				// Possibly the stride of the texture map entries?  It always seems to be 8, so it's hard to say.
	uint16_t	textureMapCount;		// The number of entries in the array at textureMapArrayOffset
	uint16_t	unknown14;
	uint16_t	matrixElementCount;		// The number of entries in the array stored at matrixOffset (often observed as 12 or 16, implying a 4x3 or 4x4 matrix)
} wmb3_material_t;

typedef struct wmb3_textureMap_s {
	uint32_t	textureTypeNameOffset;	// Offset to a null-terminated string that indicates how to use the given texture (observed values: "Albedo0", "Albedo1", "LightMap", "CubeMap0", "NormalMap0", "SpePow")
	uint32_t	textureIdentifierIndex;	// The index of the texture identifier (see: wmb3_textureIdentifier_t) to use
} wmb3_textureMap_t;

// A global texture identifier (corresponds with the identifiers given in the wta file)
typedef uint32_t wmb3_textureIdentifier_t;

/********** Mesh Names **********/

typedef struct wmb3_meshNameHeader_s {
	uint32_t	nameOffset;	// Offset to a null-terminated string
	uint32_t	unknown04;
	float		unknown08;
	float		unknown0C;
	float		unknown10;
	float		unknown14;
	float		unknown18;
	uint32_t	meshIdArrayOffset;	// 0x1C Offset to an array of shorts which indicate which meshes this names applies to (seems redundant with wmb3_meshMetadata_t)
	uint32_t	meshIdCount;
	uint32_t	unknown24;	// 0x00000000
	uint32_t	unknown28;	// 0x00000000
	uint32_t	unknown2C;	// 0x00000000
	uint32_t	unknown30;	// 0x00000000
	uint32_t	unknown34;	// Offset to an array of shorts which look like some sort of mapping (very frequently null)
	uint32_t	unknown38;	// Count
	uint32_t	unknown3C;	// 0x00000000
	uint32_t	unknown40;	// 0x00000000
	uint32_t	unknown44;	// Offset to an array of shorts which look like some sort of mapping
	uint32_t	unknown48;	// Count
} wmb3_meshNameHeader_t;
